-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, Graphics
	description: "Drawable Text"
	interface: client

class DRAWABLE_TEXT

inherit
	DRAWABLE
		end

creation
	make

feature -- Creation

	make, reset is
			-- Create/initialise text object.
		local
			black: RGB_COLOR
		do
			if text = Void then
				!!text.make (Default_text_size)
			else
				text.wipe_out
			end
			
			!!black.black
			win_color := black.to_windows

			align_left
			align_top
			
			pos_x := 0 pos_y := 0
			
			has_area := false
			area_x := 0 area_y := 0 area_xp := 0 area_yp := 0

			justification := 0
			intercharacter_spacing := 0
		end

feature {NONE} -- Attributes

	text: STRING
			-- Text

	pos_x, pos_y: INTEGER
			-- Text position

	has_area: BOOLEAN
			-- Is position a full rectangle.

	area_x, area_y, area_xp, area_yp: INTEGER
			-- Area for clipping/filling.

	win_color: INTEGER
			-- Text colour

	justification: INTEGER
			-- Justification

	intercharacter_spacing: INTEGER
			-- Intercharacter spacing

feature {NONE}

	vert_align: INTEGER
			-- Win32 vertical alignment.

	horiz_align: INTEGER
			-- Win32 horizontal alignment.
		
feature {NONE} -- Constants

	Default_text_size: INTEGER is 100

feature -- Text
	
	set_text (txt: STRING) is
			-- Set the text to draw.
		require
			valid_txt: txt /= Void
		do
			text.copy (txt)
		ensure
			done: text.is_equal (txt)
		end

feature -- Text output specifications

	set_position (pt: POINT) is
			-- Set reference point used to position the text.
		require
			valid_position: pt /= Void
		do
			pos_x := pt.x
			pos_y := pt.y
		ensure
			done: pos_x = pt.x and pos_y = pt.y
		end

	set_area (rpos: RECTANGLE) is
			-- Set area which will be filled with background colour, 
			-- assumes window device in text mode.
			-- (Default: none, text output transparent).
		require
			ok: rpos /= Void
		do
			has_area := true
			area_x := rpos.lower.x
			area_y := rpos.lower.y
			area_xp := rpos.upper.x + 1 -- Windows' rectangles...
			area_yp := rpos.upper.y + 1
		ensure
			done: has_area
			area_ok: area_x = rpos.lower.x and area_y = rpos.lower.y and area_xp = rpos.upper.x+1 and area_yp = rpos.upper.y+1
		end

	set_color (clr: COLOR) is
			-- Set text colour.
		require
			valid_color: clr /= Void
		do
			win_color := clr.to_windows
		end

	set_justification (justif: INTEGER) is
			-- Set justification. The justification value is the width 
			-- to be added to the string at the break characters.
		require
			valid: justif >= 0
		do
			justification := justif
		ensure
			done: justification = justif
		end

	set_intercharacter_spacing (ispace: INTEGER) is
			-- Set intecharacter spacing (default: 0).
		require
			valid: ispace >= 0
		do
			intercharacter_spacing := ispace
		ensure
			done: intercharacter_spacing = ispace
		end

feature -- Alignment

	align_top is
			-- Reference point aligned with the top of the bounding rectangle (default).
		do
			vert_align := wrt.Ta_top
		end

	align_baseline is
			-- Reference point aligned with the base line of the text.
		do
			vert_align := wrt.Ta_baseline
		end

	align_bottom is
			-- Reference point aligned with the bottom of the bounding rectangle.
		do
			vert_align := wrt.Ta_bottom
		end

	align_left is
			-- Reference point horizontally aligned with the left of the bounding rectangle (Default).
		do
			horiz_align := wrt.Ta_left
		end

	align_center is
			-- Reference point aligned with the centre of the bounding rectangle.
		do
			horiz_align := wrt.Ta_center
		end

	align_right is
			-- Reference point horizontally aligned with the right of the bounding rectangle.
		do
			horiz_align := wrt.Ta_right
		end

feature {GRAPHIC_DEVICE} -- Size mesurement before printing

	text_extent (dc: GRAPHIC_DEVICE): RECTANGLE is
			-- Size of the rectangle bounding the text, if it was printed now
		require
			valid_device: dc /= Void
		do
			if intercharacter_spacing /= 0 then
				wrt.win_gdi_textis (dc.handle, intercharacter_spacing)
			end
			
			internal_get_size (dc)

			!!Result.make
			Result.set_width (wrt.win_lastint1)
			Result.set_height (wrt.win_lastint2)

			if intercharacter_spacing /= 0 then
				wrt.win_gdi_textis (dc.handle, 0)
			end
		ensure
			valid_result: Result /= Void and then Result.lower.is_origin
		end

feature {GRAPHIC_DEVICE}

	draw (dc: GRAPHIC_DEVICE) is
			-- Actual drawing
		local
			countbreak: INTEGER
			breakchar: INTEGER
		do
			wrt.win_gdi_textclr (dc.handle, win_color)
			wrt.win_gdi_textalign (dc.handle, horiz_align, vert_align)

			if justification > 0 then
				breakchar := wrt.win_gdi_textmetr (dc.handle, wrt.Ftm_breakchar)
				
				countbreak := count_breaks (breakchar)

				wrt.win_gdi_textjustif (dc.handle, justification, countbreak)
			end
			
			if intercharacter_spacing /= 0 then
				wrt.win_gdi_textis (dc.handle, intercharacter_spacing)
			end

			if has_area then
				internal_ext_text_out (dc)
			else
				internal_text_out (dc)
			end

			if intercharacter_spacing /= 0 then
				wrt.win_gdi_textis (dc.handle, 0)
			end

			if justification > 0 then
				wrt.win_gdi_textjustif (dc.handle, 0, 0)
			end
		end
	
feature {NONE} -- (Internal) text functions

	internal_get_size (dc: GRAPHIC_DEVICE) is
		do
			wrt.win_gdi_textsize (dc.handle, wrt.ew_string (text).to_c)
		end

	internal_text_out (dc: GRAPHIC_DEVICE) is
		do
			wrt.win_gdi_text (dc.handle, pos_x, pos_y, wrt.ew_string (text).to_c)
		end

	internal_ext_text_out (dc: GRAPHIC_DEVICE) is
		do
			wrt.win_gdi_extext (dc.handle, pos_x, pos_y, area_x, area_y, area_xp, area_yp,
					false, true, wrt.ew_string (text).to_c)
		end

	count_breaks (breakchar: INTEGER): INTEGER is
		local
			i: INTEGER
		do
			from
				i := 1
			variant
				text.count + 1 - i
			until
				i > text.count
			loop
				if text.item (i).code = breakchar then
					Result := Result + 1
				end
					
				i := i + 1
			end
		end

invariant
	valid: text /= Void

end -- class DRAWABLE_TEXT
