-- Pylon: a foundation library; version 0.82(beta); 1998-01-04

-- 
-- Copyright (c) 1997 Franck Arnaud, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY FRANCK ARNAUD AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHORS BE LIABLE TO ANY PARTY FOR 
-- ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
-- DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE CODE.
-- 

indexing
	cluster: Pylon, Time
	description: "Date and time"
	interface: client

class P_DATE_TIME

inherit
	HASHABLE
		redefine
			is_equal, out
		end
	P_TEXT_OBJECT
		undefine
			is_equal, out
		end


creation
	make

feature -- Creation

	make is
		do
			!!date
			!!time
			internal_tz := Local_timebias
		end

feature -- Comparison

	is_equal(other : like Current) : BOOLEAN is
		do
			Result := date.is_equal(other.date) and time.is_equal(time)
					and is_local = other.is_local
					and ((not is_local) implies (timezone_bias = other.timezone_bias))
		end

feature -- Hashable

	hash_code : INTEGER is
		do
			Result := date.hash_code + time.hash_code + internal_tz
		end

feature {PUBLIC_NONE} -- Text object

	object_to_string(tm : P_TEXT_OBJECT_MANAGER) is
		do
			tm.put_integer(1)
			tm.put_integer(date.year)
			tm.put_integer(date.month)
			tm.put_integer(date.day)
			tm.put_integer(time.hour)
			tm.put_integer(time.minute)
			tm.put_integer(time.second)
			tm.put_integer(internal_tz) -- Local_timebias for local time
		end

	string_to_object(tm : P_TEXT_OBJECT_MANAGER) is
		local
			version : INTEGER
		do
			tm.get_integer
			version := tm.last_integer
			tm.get_integer
			date.set_year(tm.last_integer)
			tm.get_integer
			date.set_month(tm.last_integer)
			tm.get_integer
			date.set_day(tm.last_integer)
			tm.get_integer
			time.set_hour(tm.last_integer)
			tm.get_integer
			time.set_minute(tm.last_integer)
			tm.get_integer
			time.set_second(tm.last_integer)
			tm.get_integer
			internal_tz := tm.last_integer

			if not tm.in_success or version /= 1 then
				make
			end
		end

feature
	
	date : P_DATE
		-- Gregorian date

	time : P_TIME
		-- Time

	set_date(yyyy, mm, dd : INTEGER) is
			-- Set date.
		require
			valid: -- see DATE.set
		do
			date.set(yyyy, mm, dd)
		ensure
			done: date.is_valid
		end

	set_time(hh, mm, ss : INTEGER) is
		require
			valid: -- see TIME.set
		do
			time.set(hh, mm, ss)
		end

	set(yyyy, mo, dd : INTEGER; hh, mi, ss : INTEGER) is
		require
			valid: -- see DATE.set and TIME.set
		do
			date.set(yyyy, mo, dd)
			time.set(hh, mi, ss)
		ensure
			done: date.is_valid
		end

feature -- Status

	is_valid : BOOLEAN is
			-- Is the current object holding valid date & time?
		do
			Result := date.is_valid
		end

	is_local : BOOLEAN is
			-- Do this date and time have NO timezone information attached?
		do
			Result := internal_tz = Local_timebias
		end

feature -- Timezone

	is_utc : BOOLEAN is
			-- Is the current time zone UTC (GMT)?
		require
			has_tz: not is_local
		do
			Result := timezone_bias = 0
		ensure
			done: (Result = true) = (timezone_bias = 0)
		end

	timezone_bias : INTEGER is
		require
			has_tz: not is_local
		do
			Result := internal_tz
		end
	
	set_timezone_bias(bias : INTEGER) is
			-- Timezone bias used to convert from the current time to GMT in 
			-- minutes.
		require
			bounds: bias < 1440 and bias > - 1440 
		do
			internal_tz := bias
		ensure
			not_local: not is_local
		end
			

feature {NONE}
	
	Local_timebias : INTEGER is 10000

	internal_tz : INTEGER -- timezone bias.
	
feature -- Output

	to_iso, out : STRING is
			-- Output date and time in ISO 8601 format.
		do
			!!Result.make(0)
			Result.copy(date.to_iso)
			Result.append_character('T')
			Result.append_string(time.to_iso)

			if not is_local then
				add_iso_timezone(Result, false)
			end
		end


	to_iso_long : STRING is
			-- Output date in ISO verbose format.
		do
			!!Result.make(0)
			Result.copy(date.to_iso_long)
			Result.append_character(' ')
			Result.append_string(time.to_rfc)

			if not is_local then
				add_iso_timezone(Result, true)
			end
		end

	to_rfc : STRING is
			-- Output date & time in RFC format.
		do
			!!Result.make(0)
			Result.copy(date.to_rfc)
			Result.append_character(' ')
			Result.append_string(time.to_rfc)
			if not is_local then
				if is_utc then
					Result.append_string(" GMT")
				--else --? add timezone +-xx
				end
			end
		end

feature {NONE}
	
	add_iso_timezone(str : STRING; verbose : BOOLEAN) is
			-- Add ISO timezone info ("Z|([ ](+|-)HH[:][mm])")to (time) string 'str'.
		local
			positive_bias : INTEGER
			fmt : P_FORMAT
		do
			if is_utc then
				str.append_character('Z')
			else
				!!fmt
				if verbose then
					str.append_character(' ') -- easier to read.
				end

				if timezone_bias >= 0 then
					str.append_character('+')
					positive_bias := timezone_bias
				else
					str.append_character('-')
					positive_bias := - timezone_bias
				end
				check pos: positive_bias >= 0 end

				str.append_string(fmt.int_to_str_min(positive_bias // 60,2))
				
				if verbose then 
					if positive_bias \\ 60 /= 0 then
						str.append_character(':')
						str.append_string(fmt.int_to_str_min(positive_bias \\ 60,2))
					end
				else
					str.append_string(fmt.int_to_str_min(positive_bias \\ 60,2))
				end
			end
		end

invariant
	valid_date: date /= Void
	valid_time: time /= Void

end
