-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, Bubble
	description: "Abstract container for drawing"
	interface: client

class DRAWING

creation
	make

feature {NONE} -- Creation

	make is
			-- Init object.
		do
			!P_LINKED_LIST[BUBBLE_OBJECT]!bubbles.make
			!P_LINKED_LIST[BUBBLE_OBJECT]!frames.make
			!!area_changed.make
		end

feature {NONE} -- (Private) attributes

	bubbles: P_LIST[BUBBLE_OBJECT]
			-- List of bubbles & arrows (foreground).
	frames: P_LIST[BUBBLE_OBJECT]
			-- List of frames (background).

feature -- Result attribute

	area_changed: RECTANGLE
			-- Area of drawing changed by last operation.

	is_blank: BOOLEAN is
			-- Is the drawing blank?
		do
			Result := bubbles.is_empty and frames.is_empty
		end

feature -- Change
	
	reset is
			-- Drawing initialised.
			-- (area_changed not updated, full redraw required)
		do
			frames.reset
			bubbles.reset
		end
	
	add_object (obj: BUBBLE_OBJECT) is
			-- Add object to drawing.
		do
			if is_frame (obj) then
				frames.add (obj)
			else
				bubbles.add (obj)
			end
			area_changed.copy (obj.bounding_box)
		end
	
	remove_object (obj: BUBBLE_OBJECT) is
			-- Remove object.
		require
			ok: obj /= Void
			inside: -- has_object (obj)
		local
			it: P_ITERATOR[BUBBLE_OBJECT]
		do
			-- init area changed
			area_changed.copy (obj.bounding_box)

			frames.equality_search (obj)
			if frames.found then
				frames.remove
			else
				-- remove dependents
				from
					it := bubbles.iterator
					it.first
				until it.outside loop
					if it.item.is_dependent (obj) or it.item = obj then
						-- update redraw area
						area_changed.add (it.item.bounding_box)
						
						-- delete dependent object
						it.synchronize_container
						it.forth
						if it.outside then
							bubbles.remove
						else
							it.lock
							bubbles.remove
							it.unlock
						end
					else
						it.forth
					end
				end
			end -- not frame
		end

feature -- Search

	find_object (pt: POINT): BUBBLE_OBJECT is
			-- Find object at point.
		require
			ok: pt /= Void
		do
			-- foreground
			Result := single_find_object (bubbles.iterator, pt)

			-- background
			if Result = Void then
				Result := single_find_object (frames.iterator, pt)
			end
		ensure
			not_found: -- implies Result = Void
		end

	dependent_area (obj: BUBBLE_OBJECT): RECTANGLE is
			-- Area depending from given object.
		require
			ok_obj: obj /= Void
			exists: -- has_object (obj)
		local
			it: P_ITERATOR[BUBBLE_OBJECT]
		do
			Result := obj.bounding_box
			if not is_frame (obj) then -- FRAME objects have no dependents.
				from
					it := bubbles.iterator
					it.first
				variant it.forward_variant until it.outside loop
					if it.item.is_dependent (obj) then
						Result.add (it.item.bounding_box)
					end
					it.forth
				end
			end
		end

feature {NONE} -- Single object serial processing

	single_find_object (it: P_ITERATOR[BUBBLE_OBJECT]; pt: POINT): BUBBLE_OBJECT is
			-- Find object on a single list.
		require
			ok_point: pt /= Void
			ok_iterator: it /= Void and then it.outside
		do
			from
				it.last
			variant it.backward_variant until it.outside loop
				if it.item.is_inside (pt) then
					Result  := it.item
					it.stop
				else
					it.back
				end
			end
		end

feature -- Load/save

	load_drawing (tf: F_TEXTFILE) is
			-- Load drawing from file.
		require
			ok: tf /= Void
			open: tf.is_alive
			read: tf.is_readable
		local
			str: STRING
			cframe: FRAME
			cbubble: BUBBLE
			carrow: ARROW
			bub: P_DICTIONARY[BUBBLE,STRING]
			arr: P_LIST[ARROW]
			it: P_ITERATOR[ARROW]
			fmt: P_FORMAT
		do
			!!fmt
			-- prepare
			!!bub.make
			!P_LINKED_LIST[ARROW]!arr.make
			-- process file
			from invariant
				has_bubble_dict: bub /= Void
				has_arrow_list: arr /= Void
			until not tf.has_data or tf.has_error loop
				tf.read_string
				str := tf.last_string
				if str.count > 8 then
					if fmt.start_equal (str, "arrow") then
						!!carrow.make_store (Current)
						carrow.from_string (str.substring (6, str.count))
						arr.add (carrow)
					elseif fmt.start_equal (str, "frame") then
						!!cframe.make
						cframe.from_string (str.substring (6, str.count))
						frames.add (cframe)
					elseif fmt.start_equal (str, "bubble") then
						!!cbubble.make
						cbubble.from_string (str.substring (7, str.count))
						bub.add (cbubble, cbubble.reference_id)
						bubbles.add (cbubble)
					end
				end
			end

			-- resolve references
			from
				it := arr.iterator 
				it.first
			variant it.forward_variant until it.outside loop
				it.item.resolve (bub)
				if it.item.is_valid then
					bubbles.add (it.item)
				end
				it.forth
			end

			-- bye
			bub := Void
			arr := Void
		end

	save_drawing (tf: F_TEXTFILE) is
			-- Save drawing to file.
		require
			ok: tf /= Void
			open: tf.is_alive
			read: tf.is_writable
		do
			single_save (tf, bubbles.iterator)
			single_save (tf, frames.iterator)
		end
		
feature {NONE}
	
	single_save (tf: F_TEXTFILE; it: P_ITERATOR[BUBBLE_OBJECT]) is
			-- Save one list of objects.
		local
			ln: STRING
		do
			from
				!!ln.make (0)
				it.first
			variant it.forward_variant until it.outside loop
				ln.copy (it.item.type)
				ln.append_string (it.item.to_string)
				tf.write_string (ln)

				it.forth
			end
		end


feature {NONE}
	
	is_frame (obj: BUBBLE_OBJECT): BOOLEAN is
			-- Is this object a frame?
		require
			ok: obj /= Void
		local
			f: FRAME
		do
			f ?= obj
			Result := (f /= Void)
		end

feature -- Drawing
	
	paint (device: GRAPHIC_DEVICE) is
			-- Draw.
		require
			ok: device /= Void
			ready: device.is_ready
		do
			-- frame paint
			single_paint (device, frames.iterator)

			-- bubble paint
			single_paint (device, bubbles.iterator)
		end
	
feature -- Drawing

	single_paint (device: GRAPHIC_DEVICE; it: P_ITERATOR[BUBBLE_OBJECT]) is
			-- Draw a single object list.
		require
			ok_device: device /= Void and device.is_ready
			ok_iterator: it /= Void
		do
			from
				it.first
			variant it.forward_variant until it.outside loop
				it.item.draw (device)
				it.forth
			end
		end

end

