-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, System
	description: "Base class for asynchronous file IO"
	interface: abstract

deferred class F_ASYNCH_IO

inherit
	F_ASYNCH_STREAM

	WIN_HANDLE

feature {CREATION_NONE}
	
	make is
			-- Create asynchronous I/O object.
		do
			finalisation_register
			!!over_read.make_read (Current)
			!!over_write.make_write (Current)
			!!buffer_read.make (1)
			!!buffer_write.make (1)
			is_alive := false
		end

feature {NONE} -- Private attributes

	over_read, over_write: WIN_OVERLAPPED
	buffer_read, buffer_write: F_MEMORY
	pos_read, pos_write: INTEGER

feature {NONE}
	
	init_connect is
			-- Initialisation for `connect' implementations.
		require
			set_alive: -- is_alive set.
		do
			-- init attributes
			pos_read := 0
			pos_write := 0
			is_reading := false
			is_writing := false
			
			-- start event
			if is_alive then
				when_connected
			end
		end

feature
	
	imperative_connect is
			-- Connect, exception violation if it fails.
		do
			connect
			if not is_alive then
				wrt.raise_exception (wrt.Excep_file)
			end
		end

	disconnect is
			-- End asynchronous file operation.
			-- Note: `when_died' is called.
		do
			is_alive := false
			when_died
			dispose -- close handle
		end

feature -- Status

	is_connect_readable: BOOLEAN;
			-- Open mode: read.

	is_connect_writable: BOOLEAN;
			-- Open mode: write.

	is_reading: BOOLEAN;
			-- Asynchronous read under way.

	is_writing: BOOLEAN;
			-- Asynchronous write under way.

	is_alive: BOOLEAN;
			-- Current connection alive?

	has_data: BOOLEAN is
			-- Data waiting (connection alive for asynchronous file).
		do
			Result := is_alive and is_valid_handle
		end

	has_error: BOOLEAN is
			-- Error status (not is_alive)/
		do
			Result := not is_alive
		end

feature

	last_byte: INTEGER is
			-- Result of last read operation.
		do
			Result := buffer_read.byte (1)
		end

	read_byte is
			-- Read a byte, call when_read when byte read.
		local
			done: BOOLEAN
		do
			-- prepare overlapped
			over_read.prepare (pos_read)
			pos_read := pos_read + buffer_read.size

			-- read file
			done := wrt.win_file_asread (handle, buffer_read.to_windows, buffer_read.size, over_read.to_windows)
			if done then -- done synchronously
				when_read
			else -- failed?
				if wrt.win_lastbool1 then -- IO pending
					is_reading := true
					over_read.wait
				else -- failed
					disconnect
				end
			end
		ensure then
			writing: -- is_writing
		end

	write_byte (byte: INTEGER) is
			-- Asynchronous write byte.
			-- `when_written' called when operation terminates.
		local
			done: BOOLEAN
		do
			-- prepare buffer
			buffer_write.set_byte (1, byte)
			
			-- prepare overlapped
			over_write.prepare (pos_write)
			pos_write := pos_write + buffer_write.size

			-- write file
			done := wrt.win_file_aswrite (handle, buffer_write.to_windows, buffer_write.size, over_write.to_windows)
			if done then 
				when_written
			else -- failed?
				if wrt.win_lastbool1 then -- IO pending
					is_writing := true
					over_write.wait
				else -- failed
					disconnect
				end
			end
		end

feature {WIN_OVERLAPPED} -- Events

	internal_when_read is
		require
			reading: is_reading
		local
			got: INTEGER
		do
			got := wrt.win_file_asok (handle, over_read.to_windows)
			if got = buffer_read.size then
				is_reading := false
				when_read
			else -- failed
				disconnect
			end
		end

	internal_when_written is
		require
			writing: is_writing
		local
			sent: INTEGER
		do
			sent := wrt.win_file_asok (handle, over_write.to_windows)
			if sent = buffer_write.size then
				is_writing := false
				when_written
			else -- failed
				disconnect
			end
		end

feature -- Default writing interface --? remove from base class

	push_byte (byte: INTEGER) is
			-- Push byte onto file (queue if file busy).
		require
			alive: is_alive
			ok: byte >= 0 and byte <= Byte_maximum
		do
			if write_queue = Void then
				!!write_queue.make
			end
			
			if not is_writing then
				write_byte (byte)
			else
				write_queue.add (byte)
			end
		end

	push_string (str: STRING) is
			-- Push string onto file (queue if busy).
			-- (the string is transmitted as is, without anything, like newline, added).
		require
			ok: str /= Void
			alive: is_alive
			narrow: -- for_all n, str.item (n).code <= Byte_maximum
		local
			i, cnt: INTEGER
		do
			from
				i := 1
				cnt := str.count
			variant cnt + 1 - i until i > cnt loop
				push_byte (str.item (i).code)
				i := i + 1
			end
		end

	when_written is
			-- Previous byte written.
		do
			-- write next
			if not write_queue.is_empty then
				write_byte (write_queue.item)
				write_queue.remove
			end
		end

feature {NONE}
	
	write_queue: P_QUEUE[INTEGER]

end -- class F_ASYNCH_FILE
