-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, System
	description: "Operating System synchronous file"
	interface: client

class F_FILE

inherit
	F_STREAM

	WIN_HANDLE

creation
	make

feature {CREATION_NONE} -- Creation
	
	make is
			-- Create a file object.
		do
			finalisation_register
			!!filename.make (0)
			reset_handle
			reset_mode
			reset_share
			reset_open
		end

feature {NONE}
	
	filename: STRING
	read, write: BOOLEAN
	sharer, sharew: BOOLEAN
	createf: INTEGER

feature {NONE}
	
	random_access: BOOLEAN is
		do
			Result := false
		end

feature -- Assertions
	
	is_valid_name (str: STRING): BOOLEAN is
			-- Is the string a valid (Win32) file name?
		require
			valid: str /= Void
		local
			i: INTEGER
			ch: CHARACTER
			chcode: INTEGER
		do
			Result := true
			
			from
				i := 1
			variant
				str.count - i + 1
			until
				i > str.count
			loop
				ch := str.item (i)
				chcode := ch.code
				if (chcode >= 0 and chcode < 32) or else ch = '%"' or else
							ch = '>' or else ch = '<' or else ch = '|' then
					Result := false
				end

				i := i + 1
			end
		end

feature -- Access mode

	set_name (str: STRING) is
			-- Set the file name.
		require
			closed: not is_alive
			valid_string: str /= Void
			valid_file_name: is_valid_name (str)
		do
			filename.copy (str)
		end

	set_mode_read, reset_mode is
			-- Set mode to be read only (default).
		require
			closed: not is_alive
		do
			read := true
			write := false
		end

	set_mode_write is
			-- Set mode to write only.
		require
			closed: not is_alive
		do
			read := false
			write := true
		end

	set_mode_read_write is
			-- Set mode to be read/write.
		require
			closed: not is_alive
		do
			read := true
			write := true
		end

	set_share_read is
			-- Set sharing to read.
		require
			closed: not is_alive
		do
			sharer := true
			sharew := false
		end

	reset_share, set_share_no is
			-- Set sharing mode to no sharing (default).
		require
			closed: not is_alive
		do
			sharer := false
			sharew := false
		end

	reset_open, set_open_zero is
			-- Create new file or reset existing one.
		require
			closed: not is_alive
			relevant: -- when is_writable
		do
			createf := wrt.Create_always
		end

	set_open_new is
			-- Create new file, fail if file exists.
		require
			closed: not is_alive
			relevant: -- when is_writable
		do
			createf := wrt.Create_new
		end

	set_open_existing is
			-- Open existing file only.
		require
			closed: not is_alive
			relevant: -- when is_writable
		do
			createf := wrt.Open_existing
		end

	set_open_replace is
			-- Open existing file and truncate it, fail if there 
			-- is no existing file.
		require
			closed: not is_alive
			writable: is_writable
		do
			createf := wrt.Truncate_existing
		end

	set_open_always is
			-- Open existing file without changing it, or create 
			-- if no file exists.
		require
			closed: not is_alive
			relevant: -- when is_writable
		do
			createf := wrt.Open_always
		end

feature -- Access
	
	connect is
			-- Open the file.
		local
			cf: INTEGER
		do
			cf := createf
			if not write then
				cf := wrt.Open_existing
			end

			reset_buffer
			set_handle (wrt.win_file_open (wrt.ew_string (filename).to_c,
					cf, read, write, sharer, sharew, random_access))
		end

	imperative_connect is
			-- Open the file and raise an exception if failed.
		do
			connect
			if not is_alive then
				wrt.raise_exception (wrt.Excep_file)
			end
		end

	disconnect is
			-- Close the file 
			-- (REQUIRED in order to flush temporary buffers).
		do
			write_flush
			dispose
		end

	is_alive, is_open: BOOLEAN is
			-- Is the file stream open? (validity and openess is the same 
			-- for this kind of stream).
		do
			Result := is_valid_handle
		end

	is_readable: BOOLEAN is
		do
			Result := read
		end

	is_writable: BOOLEAN is
		do
			Result := write
		end

	has_data: BOOLEAN is
		do
			if buffer_max = 0 then
				read_buffer
			end
			
			if buffer_max /= 0 then
				Result := true
			end
		end

	has_error: BOOLEAN;
			-- Has the stream got errors?
			
feature {NONE}

	write_count, buffer_index, buffer_max: INTEGER
	write_buffer, buffer: F_MEMORY
	lastb: INTEGER

feature {NONE} -- Read cache management.

	Buffer_size: INTEGER is 2000
	
	reset_buffer is
		do
			-- read buffer
			if buffer = Void then
				!!buffer.make (Buffer_size)
			end
			buffer_index := 0
			buffer_max := 0

			-- write buffer
			if write_buffer = Void then
				!!write_buffer.make (Buffer_size)
			end
			write_count := 0
		ensure
			read_ready: buffer /= Void and then buffer.size = Buffer_size
			write_ready: write_buffer /= Void and then write_buffer.size = Buffer_size
			init_read: buffer_index = 0 and buffer_max = 0
			init_write: buffer_max = 0
		end

	read_buffer is
			-- Low-level read into buffer.
		require
			valid_index: buffer_max = 0
		do
			buffer_max := wrt.win_file_read (handle, buffer.to_windows, buffer.size)
			buffer_index := 0
		end

	write_flush is
			-- Flush write buffer.
		do
			if write_count /= 0 then
				if wrt.win_file_write (handle, write_buffer.to_windows, write_count) /= write_count then
					has_error := true
				end
				write_count := 0
			end
		ensure
			done: write_count = 0
		end

feature -- Byte
	
	last_byte: INTEGER is
		do
			Result := lastb
		end

	read_byte is
		do
			if buffer_max = 0 then
				read_buffer
			end
			
			if buffer_max /= 0 then
				buffer_index := buffer_index + 1
				lastb := buffer.byte (buffer_index)
				if buffer_index >= buffer_max then
					buffer_max := 0
					buffer_index := 0
				end
			end
		end

	write_byte (b: INTEGER) is
		do
			-- update position
			write_count := write_count + 1
			check buffer_big_enough: write_count <= write_buffer.size end

			-- update buffer
			write_buffer.set_byte (write_count, b)
			
			-- flush if full
			if write_count = write_buffer.size then
				write_flush
			end

			check not_full: write_count < write_buffer.size end
			--? has_error default?
			--? has_error := not wrt.win_file_writebyte (handle, b)
		end

feature -- Block

	write_block (memory: F_MEMORY) is
			-- Write a memory block to disk.
		require
			alive: is_alive
			writable: is_writable
		do
			if (memory.size + write_count) < write_buffer.size then
				-- put in buffer
				write_buffer.set_subset (write_count + 1, memory)
				write_count := write_count + memory.size
			else
				-- do not fit: flush & direct write
				write_flush
				if wrt.win_file_write (handle, memory.to_windows, memory.size) /= memory.size then
					has_error := true
				end
			end

			check not_full: write_count < write_buffer.size end
		end

	read_block (req_size: INTEGER) is
			-- Read a fixed size block from stream.
		require
			size_ok: req_size > 0
			alive: is_alive
			readable: is_readable
			waiting: has_data
		local
			lb_idx: INTEGER
			tmpblock: F_MEMORY
		do
			last_block := Void

			-- init read index
			lb_idx := 1
			
			-- read from cache
			if buffer_max /= 0 then
				-- copy from buffer what is available
				!!last_block.make (req_size)
				from
				invariant
					coherent_buffer: buffer_index <= buffer_max
				variant
					buffer_max + 1 - buffer_index
				until
					(buffer_index = buffer_max) or else (lb_idx > req_size)
				loop
					buffer_index := buffer_index + 1
					last_block.set_byte (lb_idx, buffer.byte (buffer_index))

					lb_idx := lb_idx + 1
				end

				-- whole of buffer has been read
				if buffer_index = buffer_max then
					reset_buffer
				end
			end

			-- read remaining of block directly 
			if lb_idx <= req_size then
				!!tmpblock.make (req_size - lb_idx + 1) -- new size (j..size)
				if wrt.win_file_read (handle, tmpblock.to_windows, tmpblock.size) /= tmpblock.size then
					-- read error
					has_error := true
					last_block := Void
				else
					-- successful read
					if lb_idx = 1 then -- nothing read from cache
						check valid_size: tmpblock.size = req_size end
						last_block := tmpblock
					else -- append to cache data
						last_block.set_subset (lb_idx, tmpblock)
					end
				end
			end
		ensure
			block_of: (not has_error) implies 
						(last_block /= Void and then last_block.size = req_size)
		end

	last_block: F_MEMORY;
			-- Last memory block read (new block for each read operation).

invariant
	valid_index: buffer_index <= buffer_max
	valid_buffer: is_alive implies buffer /= Void

end -- class F_FILE
