-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, System
	description: "Low-level memory block"
	interface: client

class F_MEMORY

inherit
	WIN_API
		redefine
			copy, is_equal
		end
	WIN_RUNTIME_FINAL
		undefine
			copy, is_equal
		end
	WIN_MEMORY
		undefine
			copy, is_equal
		end

creation
	make

feature {WIN_API} -- Attribute(s)
	
	to_windows: POINTER;
			-- Pointer.

feature {WIN_API} -- WIN_RUNTIME pointer

	to_c: POINTER is
			-- Pointer with kernel-compatible name (for WIN_RUNTIME).
		do
			Result := to_windows
		end

feature {WIN_API} -- Conversion
	
	from_windows (ptr: POINTER) is
			-- Import from external structure.
		require
			writable: is_writable
			size: -- sizeof (ptr) >= size
		local
			i: INTEGER
		do
			from
				i := size
			variant
				i
			until
				i = 0
			loop
				set_byte (i, wrt.win_getbyte (ptr, i-1))
				i := i - 1
			end
		end

feature {CREATION_NONE} -- Creation

	make (sz: INTEGER) is
			-- Create a memory block of `sz' size, in bytes.
		require
			positive: sz > 0
		do
			finalisation_register
			to_windows := wrt.win_mem_alloc (sz)
			size := sz
		ensure
			done: size = sz
		end

feature -- Finalisation

	close is
			-- Close memory object (free memory, handled 
			-- by finalisation if omitted).
		require
			alive: size /= 0
		do
			dispose
		ensure
			done: size = 0
		end

feature {NONE} -- Finalisation

	dispose is
			-- Free memory.
		do
			if size /= 0 then
				win_mem_free (to_windows)
				size := 0
			end
		end

feature -- Copy
	
	copy (other: like Current) is
			-- Copy other memory block. Other MUST be the same size 
			-- as Current.
		do
			check
				current_writable: is_writable 
				same_size: size = other.size
			end
			from_windows (other.to_windows)
		ensure then
			size: size = other.size
		end

	is_equal (other: like Current): BOOLEAN is
		local
			i: INTEGER
		do
			if other.size = size then
				from
					Result := true
					i := size
				variant
					i
				until
					i = 0 or else Result = false
				loop
					if other.byte (i) /= byte (i) then
						Result := false
					end
					i := i - 1
				end
			end -- else Result := false
		end

feature
	
	is_writable: BOOLEAN is
			-- Can this block be written to?
		do
			Result := true
		end

feature -- Size/Index
	
	size: INTEGER;
			-- Size of the memory block.

	offset (offs: INTEGER): INTEGER is
			-- Convert an offset to a position.
		require
			valid: offs >= 0 and offs < size
		do
			Result := offs + 1 
		ensure
			done: Result = offs + 1
		end

	is_index (idx: INTEGER): BOOLEAN is
			-- Is `idx' a valid index?
		do
			Result := (idx >= 1) and (idx <= size)
		ensure
			done: Result = (idx >= 1 and idx <= size)
		end

feature
	
	byte (index: INTEGER): INTEGER is
			-- Get the value of the byte at `index' position.
		require
			valid: is_index (index)
		do
			Result := wrt.win_getbyte (to_windows, index-1)
		ensure
			valid_byte: Result >= 0 and Result <= 255
		end

	--? unsigned byte.

	set_byte (index, value: INTEGER) is
			-- Set the value of the byte at current position.
		require
			writable: is_writable
			valid_index: is_index (index)
			valid_byte: value >= 0 and value <= 255
		do
			wrt.win_setbyte (to_windows, index-1, value)
		ensure
			done: byte (index) = value
		end

--? CHARACTER / Unicode / bit fields

feature -- Subset
	
	subset (start, length: INTEGER): F_MEMORY is
			-- Get a copy of a subset of the memory 
			-- block.
		require
			valid: is_index (start)
			length: length > 0
			bounds: is_index (start+length-1)
		local
			i: INTEGER
		do
			!!Result.make (length)
			from
				i := 1
			invariant
				i >= 1 and i <= length + 1
			variant
				length + 1 - i
			until
				i > length
			loop
				Result.set_byte (i, byte (start+i-1))
				i := i + 1
			end
		ensure 
			is_copy: Result /= Void
		end

	set_subset (idx: INTEGER; sub_set: F_MEMORY) is
			-- Set a block of memory inside this one.
		require
			writable: is_writable
			valid: is_index (idx)
			length: is_index (idx + sub_set.size - 1) 
		local
			i, j: INTEGER
		do
			from
				i := idx
				j := 1
			invariant
				i_j: j + idx = i + 1
			variant
				idx + sub_set.size - i
			until
				j > sub_set.size
			loop
				set_byte (i, sub_set.byte (j))
				
				i := i + 1
				j := j + 1
			end
		ensure
			done: subset (idx, sub_set.size).is_equal (sub_set)
		end

feature -- Get unsigned 16-bit WORDs.
		
	unsigned_word (idx: INTEGER): INTEGER is
			-- Convert this memory block to integer, as a 
			-- machine-order WORD
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := wrt.win_getword (subset (idx, 2).to_windows)
		ensure
			done: Result < Max_unsigned_word and Result >= 0
		end

	little_unsigned_word (idx: INTEGER): INTEGER is
			-- Like get_word, little endian order.
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := byte (idx) + 256 * byte (idx+1)
		ensure
			done: Result < Max_unsigned_word and Result >= 0
		end

	big_unsigned_word (idx: INTEGER): INTEGER is
			-- Like get_word, big endian order.
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := byte (idx) * 256 + byte (idx+1)
		ensure
			done: Result < Max_unsigned_word and Result >= 0
		end

feature -- Set unsigned 16-bit WORDs.

	set_unsigned_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_unsigned_word and value >= 0
		do
			wrt.win_setword (to_windows, idx-1, value)
		ensure
			done: value = unsigned_word (idx)
		end

	set_little_unsigned_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_unsigned_word and value >= 0
		do
			set_byte (idx, value \\ 256)
			set_byte (idx+1, value // 256)
		ensure
			done: value = little_unsigned_word (idx)
		end

	set_big_unsigned_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_unsigned_word and value >= 0
		do
			set_byte (idx, value // 256)
			set_byte (idx+1, value \\ 256)
		ensure
			done: value = big_unsigned_word (idx)
		end

feature -- Signed WORD (16 bits, 2 bytes).

	word (idx: INTEGER): INTEGER is
			-- Convert the two bytes at `idx' to integer as 
			-- a machine unsigned_word.
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := make_word_sign (unsigned_word (idx))
		ensure
			done: Result < Max_word and Result > - Max_word
		end

	little_word (idx: INTEGER): INTEGER is
			-- Little endian signed word.
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := make_word_sign (little_unsigned_word (idx))
		ensure
			done: Result < Max_word and Result > - Max_word
		end

	big_word (idx: INTEGER): INTEGER is
			-- Big endian signed word.
		require
			valid: is_index (idx)
			word: is_index (idx+1)
		do
			Result := make_word_sign (big_unsigned_word (idx))
		ensure
			done: Result < Max_word and Result > - Max_word
		end

feature -- Set 16-bit unsigned WORD.

	set_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_word and value > - Max_word
		do
			set_unsigned_word (idx, get_word_sign (value))
		ensure
			done: value = word (idx)
		end

	set_little_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_word and value > - Max_word
		do
			set_little_unsigned_word (idx, get_word_sign (value))
		ensure
			done: value = little_word (idx)
		end

	set_big_word (idx: INTEGER; value: INTEGER) is
		require
			writable: is_writable
			valid: is_index (idx) and is_index (idx+1)
			value: value < Max_word and value > - Max_word
		do
			set_big_unsigned_word (idx, get_word_sign (value))
		ensure
			done: value = big_word (idx)
		end

feature -- DWORD

	double_word (idx: INTEGER): INTEGER is
			-- Get (signed) 32-bit integer starting at`idx', 
			-- machine byte order.
		require
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			Result := wrt.win_getlong (subset (idx, 4).to_windows)
		end

	little_double_word (idx: INTEGER): INTEGER is
			-- Get little endian (signed) 32-bit integer.
		require
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			Result := subset (idx, 2).little_unsigned_word (1) + 
					Max_unsigned_word * subset (idx+2, 2).little_word (1)
		end

	big_double_word (idx: INTEGER): INTEGER is
			-- Get big endian (signed) 32-bit integer.
		require
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			Result := subset (idx+2, 2).big_unsigned_word (1) + 
					Max_unsigned_word * subset (idx, 2).big_word (1)
		end

feature -- Set DWORD

	set_double_word (idx: INTEGER; value: INTEGER) is
			-- Set (signed) 32-bit integer at `idx'.
			-- Machine byte order.
		require
			writable: is_writable
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			wrt.win_setlong (to_windows, idx-1, value)
		ensure
			done: double_word (idx) = value
		end

	set_little_double_word (idx: INTEGER; value: INTEGER) is
			-- Set little endian (signed) 32-bit integer at `idx'.
		require
			writable: is_writable
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			wrt.win_setlong_l (to_windows, idx-1, value)
		ensure
			done: little_double_word (idx) = value
		end

	set_big_double_word (idx: INTEGER; value: INTEGER) is
			-- Set big endian (signed) 32-bit integer at `idx'.
		require
			writable: is_writable
			valid: is_index (idx)
			dword: is_index (idx+3)
		do
			wrt.win_setlong_b (to_windows, idx-1, value)
		ensure
			done: big_double_word (idx) = value
		end

feature -- C String

	cstring (idx, length: INTEGER): STRING is
			-- Read system dependent string, character width is 
			-- the same as the system's CHARACTER. Scanning 
			-- terminates at first 0 or `length' whichever happens 
			-- first. NOTE: `length' is in bytes.
		require
			valid_index: is_index (idx)
			valid_length: length > 0
			valid_end: is_index (idx+length-1)
		do
			Result := byte_cstring (idx, length) --? Unicode
		end

	byte_cstring (idx, length: INTEGER): STRING is
			-- Read string with 8-bit characters, up to 
			-- a zero byte or `length'.
		require
			valid_index: is_index (idx)
			valid_length: length > 0
			valid_end: is_index (idx+length-1)
		local
			i, stop: INTEGER
		do
			!!Result.make (0)
			from
				i := idx
				stop := idx + length - 1
			variant 
				stop + 1 - i
			until
				i > stop or else byte (i) = 0
			loop
				Result.append_character (wrt.int_to_char (byte (i)))
				i := i + 1
			end
		end

	set_cstring (idx, length: INTEGER; str: STRING) is
		require
			writable: is_writable
			valid_index: is_index (idx)
			valid_length: length > 0
			valid_end: is_index (idx+length-1)
			valid_str: str /= Void
		do
			set_byte_cstring (idx, length, str)
		ensure
			done: cstring (idx, length).is_equal (str)
		end

	set_byte_cstring (idx, length: INTEGER; str: STRING) is
			-- Write string with one byte per characters, 
			-- bytes between str.count + 1 and length are 
			-- set to zero.
		require
			writable: is_writable
			valid_index: is_index (idx)
			valid_length: length > 0
			valid_end: is_index (idx+length-1)
			valid_str: str /= Void
		local
			i, str_i, stop: INTEGER
		do
			from
				i := idx
				str_i := 1
				stop := idx + length - 1
			invariant
				str_i = i - idx + 1
			variant
				stop + 1 - i
			until
				i > stop
			loop
				if str_i <= str.count then
					set_byte (i, str.item (str_i).code)
				else
					set_byte (i, 0)
				end
				i := i + 1
				str_i := str_i + 1
			end
			--? Unicode
		ensure
			done: byte_cstring (idx, length).is_equal (str)
		end
	
	-- wide_cstring (idx, length: INTEGER) STRING
	-- set_wide_cstring (idx, length: INTEGER; str: STRING)

feature -- Bit operations

	bit_and (a, b: INTEGER): INTEGER is
			-- Bitwise AND (like C '&')
		do
			Result := wrt.win_bitand (a, b)
		end

	bit_or (a, b: INTEGER): INTEGER is
			-- Bitwise OR (like C '|')
		do
			Result := wrt.win_bitor (a, b)
		end

	bit_xor (a, b: INTEGER): INTEGER is
			-- Bitwise XOR (like C '^')
		do
			Result := wrt.win_bitxor (a, b)
		end

feature {NONE}

	make_word_sign (int: INTEGER): INTEGER is
		do
			if int >= Max_word then
				Result := int - Max_unsigned_word
			else
				Result := int
			end
		end

	get_word_sign (int: INTEGER): INTEGER is
		do
			if int < 0 then
				Result := int + Max_unsigned_word
			else
				Result := int 
			end
		end

feature -- Constants

	Max_word: INTEGER is 32768 -- 2^15
	Max_unsigned_word: INTEGER is 65536 -- 2^16

invariant
	implementation: --? Integer_bits >= 32

end -- class F_MEMORY
