-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, System
	description: "System configuration registry access"
	interface: abstract

class REGISTRY_KEY

inherit
	WIN_OBJECT
		end
	WIN_API
		end
	WIN_RUNTIME_FINAL
		end

creation
	make_user, make_machine

feature {CREATION_NONE} -- Creation

	make_user is
			-- Create registry object based on current USER.
		do
			finalisation_register
			win_base := wrt.Hkey_current_user
		end

	make_machine is
			-- Create registry object based on current MACHINE.
		do
			finalisation_register
			win_base := wrt.Hkey_local_machine
		end

feature {NONE} -- Attributes
	
	win_base: INTEGER;
			-- Basic tree.
		
	new: BOOLEAN

	lastint: INTEGER
	laststring: STRING

feature {NONE}
	
	dispose is
		do
			if handle /= Invalid_handle then
				win_reg_close (handle)
				handle := Invalid_handle
			end
		end

feature -- Actions

	is_open: BOOLEAN is
			-- Is the current key open?
		do
			Result := is_valid_handle
		end

	is_new: BOOLEAN is
			-- Has the current key just been created?
		require
			open: is_open
		do
			Result := new
		end

	open (name: STRING) is
			-- Open an existing registry key.
		require
			valid_name: name /= Void
			closed: not is_open
		do
			set_handle (wrt.win_reg_open (win_base, wrt.ew_string (name).to_c))
			new := false
		ensure
			not_new: is_new = false
			may_fail: -- check is_open.
		end

	create (name: STRING) is
			-- Create a new key (or open an existing one).
		require
			valid_name: name /= Void
			closed: not is_open
		do
			set_handle (wrt.win_reg_create (win_base, wrt.ew_string (name).to_c))
			new := wrt.win_lastbool1
		ensure
			may_fail: -- check is_open.
		end

	close is
			-- Close the current key.
		require
			opened: is_open
		do
			dispose
		ensure
			closed: not is_open
		end

feature -- Utility

	standard_name (company, program, key: STRING): STRING is
			-- Return a standard key name for Software configuration.
			-- (that is "Software\<company>\<program>\<key>")
		require
			valid_company: company /= Void
			valid_program: program /= Void
			valid_key: key /= Void
		do
			!!Result.make (0)
			Result.copy ("Software\")
			Result.append_string (company)
			Result.append_character ('\')
			Result.append_string (program)
			Result.append_character ('\')
			Result.append_string (key)
		end

feature {NONE} -- Int/bool

	boolean_to_integer (in: BOOLEAN): INTEGER is
			-- Convert boolean to integer, recommended conversion for 
			-- storing Eiffel BOOLEAN values in an INTEGER registry 
			-- entry.
		do
			Result := 0
			if in = true then
				Result := 1
			end
		end

	integer_to_boolean (in: INTEGER): BOOLEAN is
			-- Convert boolean to integer. (see boolean_to_integer)
		do
			if in /= 0 then
				Result := true
			end
		end

feature -- List of values/subkeys

	values: P_LIST[STRING] is
			-- Enumerate value labels of this key.
		require
			open: is_open
		local
			ok: BOOLEAN
			idx: INTEGER
			str: STRING
		do
			!P_LINKED_LIST[STRING]!Result.make
			
			from
				ok := true
				idx := 0
			until not ok loop
				ok := wrt.win_reg_values (handle, idx)
				str := wrt.win_laststring
				if ok then
					Result.add (str)
				end
				idx := idx + 1
			end
		ensure
			is_copy: Result /= Void
		end

	subkeys: P_LIST[STRING] is
			-- Enumerate subkey names associated under this key.
		require
			open: is_open
		local
			ok: BOOLEAN
			idx: INTEGER
			str: STRING
		do
			!P_LINKED_LIST[STRING]!Result.make

			from
				ok := true
				idx := 0
			until not ok loop
				ok := wrt.win_reg_subkeys (handle, idx)
				str := wrt.win_laststring
				if ok then
					Result.add (str)
				end
				idx := idx + 1
			end
		ensure
			is_copy: Result /= Void
		end

	delete_subkey (name: STRING) is
			-- Delete subkey.
			-- Note: delete does not recurse under Windows NT.
		require
			ok: name /= Void
			exists: -- has_subkey (name)
		local
			res: BOOLEAN
		do
			res := wrt.win_reg_delete (handle, wrt.ew_string (name).to_c)
			--? do something with result
		end

feature -- Last read value.

	has_value: BOOLEAN;
			-- Was last read operation successful?

	last_string: STRING is
			-- Last read string.
		require
			ready: has_value
		do
			Result := laststring
		ensure
			done: Result /= Void
			copy: -- New string for each read (not for each call to last_string).
		end

	last_integer: INTEGER is
			-- Last read integer.
		require
			ready: has_value
		do
			Result := lastint
		end

	last_boolean: BOOLEAN is
			-- Last read boolean.
		require
			ready: has_value
		do
			Result := integer_to_boolean (lastint)
		end

--	last_binary: F_MEMORY is
--		require
--			ready: has_value
--		do
--		end


feature -- Read

	read_string (label: STRING) is
			-- Read a string key value.
		require
			valid: label /= Void
			open: is_open
		do
			has_value := false
			!!laststring.make (0)
			if wrt.win_reg_type (handle, wrt.ew_string (label).to_c) = wrt.Reg_sz then
				wrt.win_reg_getstring (handle, wrt.ew_string (label).to_c)
				laststring := wrt.win_laststring
				has_value := true
			end
		ensure
			may_fail: -- has_value implies last_string set
		end
		
	read_integer (label: STRING) is
			-- Read an integer key value.
		require
			valid: label /= Void
			open: is_open
		do
			has_value := false
			lastint := 0
			if wrt.win_reg_type (handle, wrt.ew_string (label).to_c) = wrt.Reg_dword then
				lastint := wrt.win_reg_getint (handle, wrt.ew_string (label).to_c)
				has_value := true
			end
		ensure
			may_fail: -- has_value implies last_integer set
		end

	read_boolean (label: STRING) is
			-- Read a pseudo-boolean key value (actually an integer).
		require
			valid: label /= Void
			open: is_open
		do
			read_integer (label)
		ensure
			may_fail: -- has_value implies last_boolean set
		end

--	read_binary (label: STRING) is
--			-- Read a binary key value
--		require
--			valid: label /= Void
--			open: is_open
--		do
--			has_value := false
--			if wrt.win_reg_type (handle, wrt.ew_string (label).to_c) = wrt.Reg_binary then
--				-- implement
--				has_value := true
--			end	
--		ensure
--			may_fail: -- has_value implies last_binary set
--		end

feature -- Write

	write_boolean (label: STRING; value: BOOLEAN) is
			-- Write a pseudo-boolean (actually an integer value 
			-- to the registry.
		require
			open: is_open
			valid_label: label /= Void
		do
			write_integer (label, boolean_to_integer (value))
		end

	write_integer (label: STRING; value: INTEGER) is
			-- Write an integer in the registry.
		require
			open: is_open
			valid_label: label /= Void
		do
			wrt.win_reg_putint (handle, wrt.ew_string (label).to_c, value)
		end

	write_string (label: STRING; value: STRING) is
			-- Write a string to the registry.
		require
			open: is_open
			valid_label: label /= Void
			valid_value: value /= Void
		do
			wrt.win_reg_putstring (handle, wrt.ew_string (label).to_c,
					wrt.ew_string (value).to_c)
		end

--	write_binary (label: STRING; value: F_MEMORY) is
--			-- Write a binary block
--		require
--			open: is_open
--			valid_label: label /= Void
--			valid_value: value /= Void
--		do
--		end 
	
end -- class REGISTRY_KEY
