-- Fenestra; version 0.80(beta); 1998-09-27

-- 
-- Copyright (c) 1998 Franck Arnaud, http://www.altsoft.demon.co.uk/
-- 
-- This package is distributed under the Eiffel Forum Freeware License, 
-- version 1 (see file forum.txt for full text). The main points are:
-- 
-- "Permission is hereby granted, without written agreement and without 
--  license or royalty fees, to use, copy, modify and/or distribute this 
--  package, provided that: copyright notices are retained unchanged [...]
-- 
--  Permission is hereby also granted [...] to distribute binary programs 
--  which depend on this package, provided that: if the binary program 
--  depends on a modified version of this package, you must publicly release 
--  the modified version of this package [...].
-- 
--  THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT WARRANTY [...]"
-- 

indexing
	cluster: Fenestra, Windows
	description: "Basic window"
	interface: inheritance

deferred class WINDOW

inherit
	COMMON_NOTIFIER
		end
	WIN_API
		end

feature {NONE} -- Window private attributes

	wo: WIN_WINDOW
			-- Associated handle.

	wd: WINDOW_DEVICE
			-- Private device.

	candidate_device: INTEGER
			-- Candidate device handle (when_drawn).

	has_actual_device: BOOLEAN is
			-- Is 'device' really being used?
		do
			Result := wd /= Void
		end

feature {WIN_API}
	
	handle: INTEGER is
			-- Shortcut to actual handle.
		do
			Result := wo.handle
		end
	
feature -- Window public attributes
	
	parent: WINDOW
			-- Parent window, it may be redefined by heirs in order to 
			-- communicate with parent window.

feature -- Graphic device

	device: WINDOW_DEVICE is
			-- Device used to draw in the window client area.
		do
			if wd = Void then 
				-- Create device on demand. This adds some 
				-- slight overhead for the check but saves 
				-- resources for all windows whose device is not 
				-- used. Also avoids some ugly code when the 
				-- device need be initialised early.
				!!wd.make (Current)
				if candidate_device /= wo.Invalid_handle then -- inside when_drawn event
					wd.prepare_paint (candidate_device)
				end
			end
			Result := wd
		ensure
			has_device: Result /= Void
		end

feature {NONE} -- OS handle & registration, accessed by notifier & heirs

	initialise_window is
			-- Basic initialisation for registration.
			-- (called also by STD_CONTROL which does do complete 
			-- registration as a STD_CONTROL is a limited kind of WINDOW)
		do
			-- inits
			is_close_ready := true
			candidate_device := wo.Invalid_handle
		end

	register, frozen window_register is
			-- Create the window, register with notifier, normally called 
			-- by creator, also does some initialisation work.
		require
			registered: wo.is_valid_handle
		do
			-- init
			initialise_window

			-- register current window with notifier
			notifier.register_window (Current)
		end

feature {CHILD_WINDOW} -- Child management

	child_list: P_LIST[CHILD_WINDOW] 
			-- List of child windows.

	add_child (win: CHILD_WINDOW) is
			-- Add a new child to this window.
		require
			valid_child: win /= Void
		do
			if child_list = Void then
				!P_LINKED_LIST[CHILD_WINDOW]!child_list.make
			end
			child_list.add (win)
		ensure
			has_list: child_list /= Void
		end

feature {CARET} -- Current caret (internal)

	caret: CARET

	set_caret (ct: CARET) is 
		do 
			caret := ct 
		end

feature {HELP_COMMAND}
	
	help_open: BOOLEAN

	set_help_open is
			-- Help has been used with this window
		do
			help_open := true
		end

feature {NONE} -- size utility

	make_size (x, y, xp, yp: INTEGER): RECTANGLE is
		local
			p1, p2: POINT
		do
			!!p1.make
			p1.set_xy (x, y)

			-- Windows does not include upper in window normally, hence the -1.
			!!p2.make
			p2.set_xy ((xp-1).max (0), (yp-1).max (0))

			!!Result.make
			Result.set_points (p1, p2)
		end

feature -- Window's size

	size: RECTANGLE is
			-- Return the window external size, in screen coordinates 
			-- for a popup window, relative to the client area of the 
			-- parent window for child windows.
		require
			valid: is_valid
		deferred
		ensure
			valid_result: Result /= Void
			is_copy: -- can be modified
		end

	client_size: RECTANGLE is
			-- Return the window's client area size, in client 
			-- coordinates, that is (x, y) = (0, 0)
		require
			valid: is_valid
		do
			wrt.win_clientsize (handle)
			Result := make_size (wrt.win_lastint1, wrt.win_lastint2, wrt.win_lastint3, wrt.win_lastint4)
		ensure
			is_copy: Result /= Void -- can be modified
			lower_zero: -- Result.lower.is_origin 
		end

	set_size (rect: RECTANGLE) is
			-- Set window size. The rectangle is in screen coordinates 
			-- for a popup windown and in the parent's window client area 
			-- coordinates for a child window.
		require
			valid_window: is_valid
			valid_rect: rect /= Void
		do
			-- Windows does not include upper in window normally, hence the +1.
			wrt.win_resize (handle, rect.lower.x, rect.lower.y, rect.upper.x+1, rect.upper.y+1)
		ensure
			size_set: rect.is_equal (size)
		end

	set_client_size (rect: RECTANGLE) is
			-- Set the client area size, in client area coordinates
			-- (so, (x, y) = (0, 0))
		require
			valid_window: is_valid
			valid_rect: rect /= Void
			lower_zero: rect.lower.x = 0 and rect.lower.y = 0
		do
			-- Windows does not include upper in window normally, hence the +1.
			wrt.win_resizeclient (handle, rect.lower.x, rect.lower.y, rect.upper.x+1, rect.upper.y+1)
		ensure
			size_set: rect.is_equal (client_size)
		end

feature {NONE}
	
	dynamic_child_resize: BOOLEAN
			-- Does this window dynamically resize child windows?

feature {CHILD_WINDOW} -- Dynamic resizing

	set_dynamic_child_resize (bool: BOOLEAN) is
			-- Dynamically resize child window (call when_parent_resized on
			-- each of them). This feature is called automatically when 
			-- using the dynamic size features of CHILD_WINDOWs.
		do
			dynamic_child_resize := bool
		ensure
			set: dynamic_child_resize = bool
		end
	
feature -- Window's text

	text: STRING is
			-- Get the window text, which can be the title bar for windows 
			-- with a caption, or used for other purposes.
		require
			valid: is_valid
		do
			wrt.win_gettitle (handle)
			Result := wrt.win_laststring
		ensure
			is_copy: -- can be modified
		end

	set_text (s: STRING) is
			-- Set the window text (see 'text').
		require
			valid_window: is_valid
			valid_string: s /= Void
		do
			wrt.win_title (handle, wrt.ew_string (s).to_c)
		ensure
			text_set: s.is_equal (text)
		end

feature -- Associated mouse managers

	set_left_mouse_manager (track: MOUSE_HANDLER) is
			-- Set associated mouse manager object for left mouse button.
		require
			void: -- track = Void implies no tracker
		do
			lmousetrack := track
		ensure
			keep_reference: lmousetrack = track -- keeps a copy of track
		end

	set_right_mouse_manager (track: MOUSE_HANDLER) is
			-- Set associated mouse manager object for right mouse button.
			-- (Default: none)
		require
			void: -- track = Void implies no tracker
		do
			rmousetrack := track
		ensure
			keep_reference: rmousetrack = track -- keeps a copy of track
		end


feature -- Window features and metrics

	is_valid: BOOLEAN is
			-- Is this window currently valid? If not it has been closed 
			-- or failed to be created properly.
		do
			Result := wo.is_valid_handle
		end

	is_iconic: BOOLEAN is
			-- Is this window reduced to an icon?
		require
			valid: is_valid
		do
			Result := wrt.win_isicon (handle)
		end

	is_enabled: BOOLEAN is
			-- Is the window able to receive user input?
		require
			valid: is_valid
		do
			Result := wrt.win_isenable (handle)
		end
	
	has_focus: BOOLEAN is
			-- Does the window has the keyboard focus?
		require
			valid: is_valid
		do
			Result := wrt.win_hasfocus (handle)
		end

	is_active: BOOLEAN is
			-- Is the window active?
		require
			valid: is_valid
		do
			Result := wrt.win_isactive (handle)
		end

	is_visible: BOOLEAN is
			-- Is the window visible?
		require
			valid: is_valid
		do
			Result := wrt.win_isvisible (handle)
		end

	is_created: BOOLEAN; 
			-- Has this window been created?

feature -- Close status

	is_close_ready: BOOLEAN
			-- Is the window ready to be closed.
			-- If true, calling close or a close message will 
			-- destroy the window. (Default value: true)

	set_close_ready (cl: BOOLEAN) is
			-- Set is_close_ready.
		do
			is_close_ready := cl
		ensure
			is_close_ready = cl
		end

feature -- Actions: destruction

	close is
			-- Close the window, and then destroy it if 
			-- is_close_ready is true.
		require
			valid: is_valid
		do
			when_closed
			if is_close_ready = true and then is_valid then
				-- is_valid is checked because for instance the 
				-- parent window may have been destroyed during 
				-- when_closed so this window has been destroyed too 
				-- by the time it tries to destroy itself.
				destroy
			end
		end

	destroy is
			-- Destroy the window (without calling 'when_closed').
		require
			valid: is_valid
		do
			wrt.win_destroy (handle)
		ensure
			invalid: -- not is_valid after when_destroyed
		end

feature -- Actions: change the way the window is displayed

	show is
			-- Show the window and redraw its client area. 
			-- Can be used to unhide, or restore a window.
		require
			valid: is_valid
		do
			wrt.win_show (handle, wrt.Sw_show)
		ensure
			visible: -- is_visible unless not parent.is_visible
		end

	hide is
			-- Hide the window (unhide with show).
		require
			valid: is_valid
		do
			wrt.win_show (handle, wrt.Sw_hide)
		ensure
			hidden: not is_visible
		end

feature -- Actions: activation, focus, etc

	activate is
			-- Activate the current window (make it top).
		require
			valid: is_valid
		do
			wrt.win_activate (handle)
		ensure
			active: is_active
		end

	focus is
			-- Set keyboard focus to this window.
		require
			valid: is_valid
		do
			wrt.win_focus (handle)
		ensure
			focus: -- has_focus unless immediatly given to someone else
		end

	enable is
			-- Allow the user to use the window.
		require
			valid: is_valid
		do
			wrt.win_enable (handle)
		ensure
			enabled: is_enabled = true
		end

	disable is
			-- Prevent the user from using the window.
		require
			valid: is_valid
		do
			wrt.win_disable (handle)
		ensure
			disabled: is_enabled = false
		end

feature -- Painting the window

	paint is
			-- Force a redraw of the client area of the window.
		require
			valid: is_valid
		do
			wrt.win_update (handle)
		end

	paint_rectangle (rect: RECTANGLE) is
			-- Force a redraw of the specified rectangle of the client area.
		require
			valid_window: is_valid
			valid_rectangle: rect /= Void
		do
			wrt.win_update_rect (handle, rect.lower.x, rect.lower.y, rect.upper.x+1, rect.upper.y+1)
		end

	horizontal_scroll (dx: INTEGER) is
			-- Scroll horizontally the client area of the window of 'dx', 
			-- and repaint the scrolled part.
		require
			valid: is_valid
		do
			wrt.win_scroll (handle, dx, 0)
		end

	vertical_scroll (dy: INTEGER) is
			-- Scroll vertically the client area of the window of 'dy', 
			-- and repaint the scrolled part.
		require
			valid: is_valid
		do
			wrt.win_scroll (handle, 0, dy)
		end
	
feature {PUBLIC_NONE} -- Windows events: Creation & Destruction

	when_created is
			-- Window created, but not yet displayed. This is the first 
			-- event.
		require
			not_created: not is_created
		do
		ensure
			created: -- system sets is_created to true after this event.
		end

	when_closed is
			-- Window is going to be closed. (if is_closed_ready = true)
		do
		end

feature {PUBLIC_NONE} -- Windows Events: Size & redraw

	when_resized is 
			-- Window has been resized (get the new coordinates 
			-- from 'size' or 'client_size').
		do
		end

	when_drawn is
			-- Window content is to be drawn.
		require
			device_ok: -- device.is_ready
		do
		end

feature {PUBLIC_NONE} -- Windows events: Mouse

	when_mouse_clicked (point: POINT) is
			-- Primary (usually left) mouse button has 
			-- been clicked.
		require
			has_point: point /= Void
			is_copy: -- `point' can be modified.
		do
		end

	when_mouse_doubleclicked (point: POINT) is
			-- Primary (usually left) mouse button has been 
			-- double clicked.
		require
			has_point: point /= Void
			is_copy: -- `point' can be modified.
		do
		end

	when_mouse_rclicked (point: POINT) is
			-- Secondary (usually right) mouse button clicked.
			-- Default behaviour: send the event to the left button.
		require
			has_point: point /= Void
			is_copy: -- `point' can be modified.
		do
			when_mouse_clicked (point)
		end

	when_mouse_rdoubleclicked (point: POINT) is
			-- Secondary (usually right) mouse button double clicked.
			-- Default behaviour: send the event to the left button.
		require
			has_point: point /= Void
			is_copy: -- `point' can be modified.
		do
			when_mouse_doubleclicked (point)
		end

feature {PUBLIC_NONE} -- Windows events: keyboard

	when_character_entered (char: CHARACTER) is
			-- A character has been entered by the user.
		do
		end

	when_key_pressed (key: VIRTUAL_KEY) is
			-- The user pressed a control key (or a normal key, but 
			-- 'when_character_entered' should be used to process 
			-- characters).
		require
			has_key: key /= Void
			is_copy: -- `key' can be modified.
		do
		end

	when_focus_given is
			-- The current window has been given keyboard focus.
		do
		end

	when_focus_lost is
			-- The current window has lost keyboard focus.
		do
		end

feature {PUBLIC_NONE} -- Windows events: misc

	when_object (object: ANY) is
			-- An eiffel object has been sent to this window.
		do
		end

feature {NONE} --? publicise
	
	when_key_released (vkey: VIRTUAL_KEY) is
			-- Key up.
		require
			valid: vkey /= Void
			is_copy: -- `vkey' may be modified.
		do
		end

feature {NONE} -- Windows events: Mouse support

	Mouse_left, Mouse_right: INTEGER is unique

	rmousetrack, lmousetrack: MOUSE_HANDLER
			-- Current mouse tracker.

feature {NONE} -- (Internal) mouse events processing

	when_mousedown (button: INTEGER ; p: POINT) is
			-- Low-level mouse management: button down.
		require
			button_ok: button = Mouse_left or button = Mouse_right
		do
			-- simple click
			inspect
				button
			when Mouse_left then
				when_mouse_clicked (p)
				if lmousetrack /= Void then
					lmousetrack.internal_when_down
				end
			when Mouse_right then
				when_mouse_rclicked (p)
				if rmousetrack /= Void then
					rmousetrack.internal_when_down
				end
			end
		end

	when_mouseup (button: INTEGER) is
			-- Low level mouse management: button up.
		require
			button_ok: button = Mouse_left or button = Mouse_right
		do
			-- up
			inspect
				button
			when Mouse_left then
				if lmousetrack /= Void then
					lmousetrack.internal_when_up
				end
			when Mouse_right then
				if rmousetrack /= Void then
					rmousetrack.internal_when_up
				end
			end
		end

	when_mousemoved (p: POINT) is
			-- Low level mouse management: mouse moved. 
		do
			if rmousetrack /= Void then
				rmousetrack.internal_when_moved (p)
			end
			if lmousetrack /= Void then
				lmousetrack.internal_when_moved (p)
			end
		end

feature {NONE}
	
	when_palette_changed is
		do
			if has_actual_device and then device.has_palette then
				device.prepare
				device.realize_palette
				paint
				device.release
			end
			-- API: we could return boolean when processing...
		end

feature {NONE}
	
	internal_when_drawn is
		local
			paintstruct: F_MEMORY
			device_handle: INTEGER
		do
			check
				not_ready: has_actual_device implies (not device.is_ready)
				candidate_init: candidate_device = wo.Invalid_handle
			end
			
			!!paintstruct.make (wrt.Paintstruct_size)
			device_handle := wrt.win_beginpaint (handle, paintstruct.to_windows)
			
			if has_actual_device then -- init device only if used
				device.prepare_paint (device_handle)
			else
				candidate_device := device_handle
			end
			
			when_drawn
			
			wrt.win_endpaint (handle, paintstruct.to_windows)
			candidate_device := wo.Invalid_handle
			
			if has_actual_device then
				device.release_paint
			end
		end

	when_destroyed, frozen standard_when_destroyed is
			-- Window destroyed, unregister with notifier (last callback).
		local
			dummy: BOOLEAN
		do
			-- close associated helpfile if one.
			if help_open = true then
				dummy := wrt.win_help_int (handle, wrt.ew_string ("").to_c, wrt.Help_quit, 0)
			end
			check valid_self: is_valid end
			notifier.unregister_window (Current)
			wo.reset_handle
		ensure
			invalid: not is_valid
			unregister: -- UNREGISTER THE WINDOW from NOTIFIER
		end

	internal_when_resized is
			-- If redefining and using dynamic child window resizing, call 
			-- 'standard_when_resized'.
		local
			it: P_ITERATOR[CHILD_WINDOW]
		do
			if dynamic_child_resize then
				check
					has_children: child_list /= Void
				end
				from
					it := child_list.iterator
					it.first
				variant
					it.forward_variant
				until
					it.outside
				loop
					it.item.when_parent_resized
					it.forth
				end
			end
			when_resized
		end

feature {NONE} -- (Internal) OVERLAPPED_WINDOW

	when_menu_selected (id: INTEGER) is
			-- for windows of type OVERLAPPED_WINDOW only.
		require
			redefined_by_heirs: false
		do
		end

	when_menu_prepared (hdle: INTEGER) is
			-- for windows of type OVERLAPPED_WINDOW only.
		require
			redefined_by_heirs: false
		do
		end

	when_menu_touched (id: INTEGER; reset: BOOLEAN) is
			-- for windows of type OVERLAPPED_WINDOW only.
		require
			redefined_by_heirs: false
		do
		end

feature {NONE} -- (Internal) APPLICATION

	internal_when_drop (drophandle: INTEGER) is
			-- APPLICATION only event
		require
			redefined_by_heirs: false
		do
		end

feature {NONE} -- (Internal) CHILD_WINDOW event(s)

	when_hscroll, when_vscroll (code, position: INTEGER) is
		require
			redefined_by_heirs: false
		do
		end

feature {NOTIFIER} -- Internal implementation of event processing.

	internal_when_object (obj: ANY) is
		do
			when_object (obj)
		end

	internal_when_created is
		require
			not_created: not is_created
		do
			when_created
			is_created := true
			show
		end

feature {WIN_EVENT} -- Low level event processing

	when_event (event: WIN_EVENT) is
			-- Process event from notifier.
		require
			valid_event: event /= Void
		local
			code: INTEGER
		do
			code := event.code
			
			-- General events
			if code = wrt.Wm_close then
				close
			elseif code = wrt.Wm_destroy then
				when_destroyed
			elseif code = wrt.Wm_size and then event.int1 /= wrt.Size_minimized then
				internal_when_resized
			elseif code = wrt.Wm_paint then
				internal_when_drawn
			
			-- Mouse events
			elseif code = wrt.Wm_lbuttondown then
				when_mousedown (Mouse_left, event.point)
			elseif code = wrt.Wm_lbuttonup then
				when_mouseup (Mouse_left)
			elseif code = wrt.Wm_lbuttondblclk then
				when_mouse_doubleclicked (event.point)
			elseif code = wrt.Wm_rbuttondown then
				when_mousedown (Mouse_right, event.point)
			elseif code = wrt.Wm_rbuttonup then
				when_mouseup (Mouse_right)
			elseif code = wrt.Wm_rbuttondblclk then
				when_mouse_rdoubleclicked (event.point)
			elseif code = wrt.Wm_mousemove then
				when_mousemoved (event.point)
			
			-- Keyboard events
			elseif code = wrt.Wm_char then
				when_character_entered (event.char)
			elseif code = wrt.Wm_keydown then
				when_key_pressed (event.vkey)
			elseif code = wrt.Wm_keyup then
				when_key_released (event.vkey)
			elseif code = wrt.Wm_setfocus then
				when_focus_given
			elseif code = wrt.Wm_killfocus then
				when_focus_lost
			
			-- Scrollbar
			elseif code = wrt.Wm_vscroll then -- built-in child window vertical scrollbar
				when_vscroll (event.int1, event.int2) -- code, pos
			elseif code = wrt.Wm_hscroll then -- built-in child window horizontal scrollbar
				when_hscroll (event.int1, event.int2) -- code, pos
							
			-- Misc
			elseif code = wrt.Wm_command then -- menu
				when_menu_selected (event.int1) -- menuid
			elseif code = wrt.Wm_initmenu or code = wrt.Wm_initmenupopup then
				when_menu_prepared (event.int1) -- menu handle
			elseif code = wrt.Wm_menuselect then
				when_menu_touched (event.int1, event.int2 = 1)
			elseif code = wrt.Wm_dropfiles then
				internal_when_drop (event.int1) -- drophandle
			elseif code = wrt.Wm_querynewpalette then
				when_palette_changed
			else
				-- Don't process
			end
		end

end -- class WINDOW
